"""Build (DataIntegration) project commands for the cmem command line interface."""

import os
import pathlib
import shutil
import tempfile
from zipfile import ZipFile

import click
from click import ClickException, UsageError
from cmem.cmempy.config import get_di_api_endpoint
from cmem.cmempy.plugins.marshalling import (
    get_extension_by_plugin,
    get_marshalling_plugins,
)
from cmem.cmempy.workspace.projects.export_ import export_project
from cmem.cmempy.workspace.projects.import_ import (
    import_from_upload_start,
    import_from_upload_status,
    upload_project,
)
from cmem.cmempy.workspace.projects.project import (
    create_project_with_transformation,
    delete_project,
    get_failed_tasks_report,
    get_projects,
    make_new_project_with_metadata,
    reload_project,
)
from jinja2 import Template

from cmem_cmemc import completion
from cmem_cmemc.command import CmemcCommand
from cmem_cmemc.command_group import CmemcGroup
from cmem_cmemc.commands.variable import variable
from cmem_cmemc.context import ApplicationContext
from cmem_cmemc.parameter_types.path import ClickSmartPath
from cmem_cmemc.smart_path import SmartPath as Path


def _validate_projects_to_process(project_ids: tuple[str], all_flag: bool) -> list[str]:
    """Return a list project IDs which will be processed.

    list of IDs is without duplicates, and validated if they exist
    """
    if project_ids == () and not all_flag:
        raise UsageError(
            "Either specify at least one project ID "
            "or use the --all option to process over all projects."
        )
    projects_to_process = list(project_ids)
    all_projects = [_["name"] for _ in get_projects()]
    if all_flag:
        # in case --all is given, a list of project is fetched
        projects_to_process = all_projects
    # avoid double removal
    projects_to_process = list(set(projects_to_process))

    # test if one of the projects does NOT exist
    for _ in projects_to_process:
        if _ not in all_projects:
            raise ClickException(f"Project {_} does not exist.")
    return projects_to_process


def _show_type_list(app: ApplicationContext) -> None:
    """Output the list of project export types.

    Internally this is named marshalling plugin.

    Args:
    ----
        app (ApplicationContext): the click cli app context.

    """
    types = get_marshalling_plugins()
    table = []
    for _ in types:
        id_ = _["id"]
        label = _["label"]
        description = _["description"].partition("\n")[0]
        row = [
            id_,
            f"{label}: {description}",
        ]
        table.append(row)
    app.echo_info_table(table, headers=["Export Type", "Description"], sort_column=1)


@click.command(cls=CmemcCommand, name="open")
@click.argument(
    "project_ids", nargs=-1, required=True, type=click.STRING, shell_complete=completion.project_ids
)
@click.pass_obj
def open_command(app: ApplicationContext, project_ids: tuple[str]) -> None:
    """Open projects in the browser.

    With this command, you can open a project in the workspace in
    your browser to change them.

    The command accepts multiple project IDs which results in
    opening multiple browser tabs.
    """
    projects = get_projects()
    for _ in project_ids:
        if _ not in (p["name"] for p in projects):
            raise ClickException(f"Project '{_}' not found.")
        open_project_uri = f"{get_di_api_endpoint()}/workbench/projects/{_}"
        app.echo_debug(f"Open {_}: {open_project_uri}")
        click.launch(open_project_uri)


@click.command(cls=CmemcCommand, name="list")
@click.option("--raw", is_flag=True, help="Outputs raw JSON.")
@click.option(
    "--id-only",
    is_flag=True,
    help="Lists only project identifier and no labels or other "
    "metadata. This is useful for piping the IDs into other commands.",
)
@click.pass_obj
def list_command(app: ApplicationContext, raw: bool, id_only: bool) -> None:
    """List available projects.

    Outputs a list of project IDs which can be used as reference for
    the project create, delete, export and import commands.
    """
    projects = get_projects()
    if raw:
        app.echo_info_json(projects)
        return
    if id_only:
        for _ in sorted(projects, key=lambda k: k["name"].lower()):
            app.echo_result(_["name"])
        return
    # output a user table
    table = []
    for _ in projects:
        row = [
            _["name"],
            _["metaData"]["label"],
        ]
        table.append(row)
    app.echo_info_table(
        table,
        headers=["Project ID", "Label"],
        sort_column=1,
        empty_table_message="No projects found. "
        "Use the `project create` command to create a new project.",
    )


@click.command(cls=CmemcCommand, name="delete")
@click.option(
    "-a",
    "--all",
    "all_",
    is_flag=True,
    help="Delete all projects. " "This is a dangerous option, so use it with care.",
)
@click.argument("project_ids", nargs=-1, type=click.STRING, shell_complete=completion.project_ids)
@click.pass_obj
def delete_command(app: ApplicationContext, all_: bool, project_ids: tuple[str]) -> None:
    """Delete projects.

    This command deletes existing data integration projects from Corporate
    Memory.

    Warning: Projects will be deleted without prompting!

    Note: Projects can be listed with the `project list` command.
    """
    projects_to_delete = _validate_projects_to_process(project_ids=project_ids, all_flag=all_)
    count = len(projects_to_delete)
    for current, project_id in enumerate(projects_to_delete, start=1):
        app.echo_info(f"Delete project {current}/{count}: {project_id} ... ", nl=False)
        delete_project(project_id)
        app.echo_success("done")


@click.command(cls=CmemcCommand, name="create")
@click.argument("project_ids", nargs=-1, required=True, type=click.STRING)
@click.option(
    "--from-transformation",
    nargs=1,
    shell_complete=completion.transformation_task_ids,
    required=False,
    help=(
        "This option can be used to explicitly create the link specification, "
        "which is internally executed when using the mapping suggestion of "
        "a transformation task. You need the task ID of the transformation task."
    ),
)
@click.option(
    "--label",
    "labels",
    multiple=True,
    help="Give the label of the project. You can give more than one label if you"
    " create more than one project.",
)
@click.option(
    "--description",
    "descriptions",
    multiple=True,
    help="Give the description of the project. You can give more than one description"
    " if you create more than one project.",
)
@click.pass_obj
def create_command(
    app: ApplicationContext,
    project_ids: tuple[str],
    from_transformation: str,
    labels: tuple[str],
    descriptions: tuple[str],
) -> None:
    """Create projects.

    This command creates one or more new projects.
    Existing projects will not be overwritten.

    Note: Projects can be listed by using the `project list` command.
    """
    if from_transformation and len(project_ids) > 1:
        raise UsageError(
            "By using --from-transformation,"
            " the project ID parameter is limited to a single project ID."
        )

    all_projects = [_["name"] for _ in get_projects()]
    for project_id in project_ids:
        if project_id in all_projects:
            raise ClickException(f"Project {project_id} already exists.")

    if from_transformation:
        transformation_parts = from_transformation.split(":")
        transformation_project, transformation_task_id = (
            transformation_parts[0],
            transformation_parts[1],
        )
        app.echo_info(
            f"Create new project {project_ids[0]} from transformation"
            f" {transformation_task_id} ... ",
            nl=False,
        )
        create_project_with_transformation(
            transform_project_id=transformation_project,
            transform_task_id=transformation_task_id,
            matching_link_spec_project_id=project_ids[0],
            matching_link_spec_id=f"linking_{transformation_task_id.split('_')[-1]}",
        )
        app.echo_success("done")
        return

    count = len(project_ids)
    current = 1
    if len(labels) > 0 and len(labels) != count:
        raise UsageError(
            "Either give labels for all projects or for no project."
            f" Got {len(labels)} labels but {count} projects."
        )
    if len(descriptions) > 0 and len(descriptions) != count:
        raise UsageError(
            "Either give descriptions for all projects or for no project."
            f" Got {len(descriptions)} descriptions but {count} projects."
        )

    for project_id in project_ids:
        try:
            label = labels[current - 1]
        except IndexError:
            label = project_id
        try:
            description = descriptions[current - 1]
        except IndexError:
            description = ""
        app.echo_info(f"Create new project {current}/{count}: {project_id} ... ", nl=False)
        app.echo_debug(get_projects())
        make_new_project_with_metadata(project_id, label=label, description=description)
        app.echo_success("done")
        current = current + 1


# pylint: disable=too-many-arguments,too-many-locals
@click.command(cls=CmemcCommand, name="export")
@click.option(
    "-a",
    "--all",
    "all_",
    is_flag=True,
    help="Export all projects.",
)
@click.option(
    "-o",
    "--overwrite",
    is_flag=True,
    hidden=True,
)
@click.option(
    "--replace",
    is_flag=True,
    help="Replace existing files. This is a dangerous option, so use it with care.",
)
@click.option(
    "--output-dir",
    default=".",
    show_default=True,
    type=ClickSmartPath(writable=True, file_okay=False),
    help="The base directory, where the project files will be created. "
    "If this directory does not exist, it will be silently created.",
)
@click.option(
    "--type",
    "marshalling_plugin",
    default="xmlZip",
    show_default=True,
    type=click.STRING,
    shell_complete=completion.marshalling_plugins,
    help="Type of the exported project file(s). Use the --help-types option "
    "or tab completion to see a list of possible types.",
)
@click.option(
    "--filename-template",
    "-t",
    "template",
    default="{{date}}-{{connection}}-{{id}}.project",
    show_default=True,
    type=click.STRING,
    shell_complete=completion.project_export_templates,
    help="Template for the export file name(s). Possible placeholders are (Jinja2): "
    "{{id}} (the project ID), "
    "{{connection}} (from the --connection option) and "
    "{{date}} (the current date as YYYY-MM-DD). "
    "The file suffix will be appended. Needed directories will be created.",
)
@click.option(
    "--extract",
    is_flag=True,
    help="Export projects to a directory structure instead of a ZIP archive. "
    "Note that the --filename-template "
    "option is ignored here. Instead, a sub-directory per exported "
    "project is created under the output directory. "
    "Also note that not all export types are extractable.",
)
@click.option("--help-types", is_flag=True, help="Lists all possible export types.")
@click.argument("project_ids", nargs=-1, type=click.STRING, shell_complete=completion.project_ids)
@click.pass_obj
def export_command(  # noqa: PLR0913
    app: ApplicationContext,
    all_: bool,
    project_ids: tuple[str],
    overwrite: bool,
    replace: bool,
    marshalling_plugin: str,
    template: str,
    output_dir: str,
    extract: bool,
    help_types: bool,
) -> None:
    """Export projects to files.

    Projects can be exported with different export formats.
    The default type is a zip archive which includes metadata as well
    as dataset resources.
    If more than one project is exported, a file is created for each project.
    By default, these files are created in the current directory with a
    descriptive name (see --template option default).

    Note: Projects can be listed by using the `project list` command.

    You can use the template string to create subdirectories.

    Example: cmemc config list | parallel -I% cmemc -c % project export --all
    -t "dump/{{connection}}/{{date}}-{{id}}.project"
    """
    if overwrite:
        replace = overwrite
        app.echo_warning(
            "The option --overwrite is deprecated and will be removed with the next major release."
            " Please use the --replace option instead."
        )

    if help_types:
        _show_type_list(app)
        return

    extractable_types = ("xmlZip", "xmlZipWithoutResources")
    if extract and marshalling_plugin not in extractable_types:
        raise UsageError(
            f"The export type {marshalling_plugin} can not be extracted. "
            f"Use one of {extractable_types}."
        )

    projects_to_export = _validate_projects_to_process(project_ids=project_ids, all_flag=all_)
    count = len(projects_to_export)
    template_data = app.get_template_data()
    for current, project_id in enumerate(projects_to_export, start=1):
        # prepare the template data and prepare target path name
        template_data.update(id=project_id)
        if extract:
            # this name is only used of display
            # the ZIP has first level directory
            local_name = project_id
        else:
            local_name = (
                Template(template).render(template_data)
                + "."
                + get_extension_by_plugin(marshalling_plugin)
            )
        # join with given output directory and normalize full path
        export_path = os.path.normpath(str(Path(output_dir) / local_name))

        app.echo_info(
            f"Export project {current}/{count}: " f"{project_id} to {export_path} ... ", nl=False
        )
        if Path(export_path).exists() and replace is not True:
            app.echo_error("target file or directory exists")
            continue

        Path(export_path).parent.mkdir(exist_ok=True, parents=True)

        if extract:
            export_data = export_project(project_id, marshalling_plugin)
            # do the export to a temp file and extract it afterward
            with tempfile.NamedTemporaryFile() as tmp_file:
                app.echo_debug(f"Temporary file is {tmp_file.name}")
                tmp_file.write(export_data)
                shutil.rmtree(export_path, ignore_errors=True)
                with ZipFile(tmp_file, "r") as _:
                    _.extractall(output_dir)
        else:
            export_data = export_project(project_id, marshalling_plugin)
            # create parent directory
            Path(export_path).parent.absolute().mkdir(exist_ok=True)
            with Path(export_path).open(mode="wb") as _:
                _.write(export_data)

        app.echo_success("done")


@click.command(cls=CmemcCommand, name="import")
@click.argument(
    "path",
    shell_complete=completion.project_files,
    type=ClickSmartPath(
        allow_dash=False, dir_okay=True, readable=True, exists=True, remote_okay=True
    ),
)
@click.argument(
    "project_id",
    type=click.STRING,
    required=False,
    default="",
    shell_complete=completion.project_ids,
)
@click.option(
    "-o",
    "--overwrite",
    is_flag=True,
    hidden=True,
)
@click.option(
    "--replace",
    is_flag=True,
    help="Replace an existing project. This is a dangerous option, so use it with care.",
)
@click.pass_obj
def import_command(
    app: ApplicationContext, path: str, project_id: str, overwrite: bool, replace: bool
) -> None:
    """Import a project from a file or directory.

    Example: cmemc project import my_project.zip my_project
    """
    if overwrite:
        replace = overwrite
        app.echo_warning(
            "The option --overwrite is deprecated and will be removed with the next major release."
            " Please use the --replace option instead."
        )

    all_projects = get_projects()
    if project_id and not replace and project_id in ([_["name"] for _ in all_projects]):
        raise ClickException(f"Project {project_id} is already there.")

    if Path(path).is_dir():
        if not (Path(path) / "config.xml").is_file():
            # fail early if directory is not an export
            raise ClickException(f"Directory {path} seems not to be a export directory.")

        app.echo_info(f"Import directory {path} to project {project_id} ... ", nl=False)
        # in case of a directory, we zip it to a temp file
        app.echo_info("zipping ... ", nl=False)
        with tempfile.NamedTemporaryFile() as _:
            shutil.make_archive(
                _.name,
                "zip",
                base_dir=pathlib.Path(path).name,
                root_dir=str(Path(path).parent.absolute()),
            )
            # make_archive adds a .zip automatically ...
            uploaded_file = _.name + ".zip"
            app.echo_debug(f"Uploaded file is {uploaded_file}")
    else:
        app.echo_info(f"Import file {path} to project {project_id} ... ", nl=False)
        with tempfile.NamedTemporaryFile(suffix=".zip", delete=False) as _:
            with ClickSmartPath.open(path) as _buffer:
                _.write(_buffer.read())
            uploaded_file = _.name

    # upload file and get validation report
    validation_response = upload_project(uploaded_file)
    # Remove the temporary file
    pathlib.Path.unlink(pathlib.Path(uploaded_file))
    if "errorMessage" in validation_response:
        raise ClickException(validation_response["errorMessage"])
    import_id = validation_response["projectImportId"]

    # get project_id from response if not given as parameter
    if not project_id:
        project_id = validation_response["projectId"]

    app.echo_debug(f"File {uploaded_file} uploaded: {validation_response}")

    # start import of project from upload using import ID as a reference
    # this fails if project_id already exists
    import_from_upload_start(import_id=import_id, project_id=project_id, overwrite_existing=replace)

    # loop until "success" boolean is in status response
    status = import_from_upload_status(import_id)
    while "success" not in status:
        status = import_from_upload_status(import_id)
    if status["success"] is True:
        app.echo_success("done")
        # output warnings in case there are failed tasks errors
        for _ in get_failed_tasks_report(project_id):
            app.echo_warning(_["errorMessage"])
    else:
        app.echo_error(" error")
    app.echo_debug(f"last import status: {status}")


@click.command(cls=CmemcCommand, name="reload")
@click.option("-a", "--all", "all_", is_flag=True, help="Reload all projects")
@click.argument("project_ids", nargs=-1, type=click.STRING, shell_complete=completion.project_ids)
@click.pass_obj
def reload_command(app: ApplicationContext, all_: bool, project_ids: tuple[str]) -> None:
    """Reload projects from the workspace provider.

    This command reloads all tasks of a project from the workspace provider.
    This is similar to the `workspace reload` command, but for a
    single project only.

    Note: You need this in case you changed project data externally or loaded
    a project which uses plugins which are not installed yet.
    In this case, install the plugin(s) and reload the project afterward.

    Warning: Depending on the size your datasets esp. your Knowledge Graphs,
    reloading a project can take a long time to re-create the path caches.
    """
    projects_to_reload = _validate_projects_to_process(project_ids=project_ids, all_flag=all_)
    count = len(projects_to_reload)
    for current, project_id in enumerate(projects_to_reload, start=1):
        app.echo_info(f"Reload project {current}/{count}: {project_id} ... ", nl=False)
        reload_project(project_id)
        app.echo_success("done")


@click.group(cls=CmemcGroup)
def project() -> CmemcGroup:  # type: ignore[empty-body]
    """List, import, export, create, delete or open projects.

    Projects are identified by a PROJECT_ID.

    Note: To get a list of existing projects, execute the `project list`
    command or use tab-completion.
    """


project.add_command(open_command)
project.add_command(list_command)
project.add_command(export_command)
project.add_command(import_command)
project.add_command(delete_command)
project.add_command(create_command)
project.add_command(reload_command)
project.add_command(variable)



"""graph validation command group"""

import json
import sys
import time
from collections import Counter
from datetime import datetime, timezone
from pathlib import Path

import click
import timeago
from click import Context, UsageError
from click.shell_completion import CompletionItem
from cmem.cmempy.config import get_cmem_base_uri
from cmem.cmempy.dp.shacl import validation
from junit_xml import TestCase, TestSuite, to_xml_report_string
from rich.progress import Progress, SpinnerColumn, TaskID, TimeElapsedColumn

from cmem_cmemc import completion
from cmem_cmemc.command import CmemcCommand
from cmem_cmemc.command_group import CmemcGroup
from cmem_cmemc.completion import finalize_completion
from cmem_cmemc.context import ApplicationContext
from cmem_cmemc.exceptions import ServerError
from cmem_cmemc.object_list import (
    DirectListPropertyFilter,
    DirectValuePropertyFilter,
    ObjectList,
    compare_int_greater_than,
    transform_lower,
)
from cmem_cmemc.string_processor import GraphLink, ResourceLink, TimeAgo
from cmem_cmemc.title_helper import TitleHelper
from cmem_cmemc.utils import get_query_text, struct_to_table


def _reports_to_junit(reports: list[dict]) -> str:
    """Create a jUnit XML document from a list of report dictionaries"""
    test_suites: list[TestSuite] = []

    for report in reports:
        test_cases: list[TestCase] = []
        context_graph = report["contextGraphIri"]
        shape_graph = report["shapeGraphIri"]
        time_elapsed = (report["executionFinished"] - report["executionStarted"]) / 1000
        violations: dict[str, list[dict]] = {}
        for resource in sorted(report["resources"]):
            # get a list of all tested resources
            violations[resource] = []
        average_elapsed_sec = time_elapsed / len(violations)
        for result in report["results"]:
            # collection violations per resource
            resource_iri = result["resourceIri"]
            violations[resource_iri] = result["violations"]
        for resource in violations:
            # create on test case per resource
            resource_violations = violations[resource]
            violations_count = len(violations[resource])
            constraints = Counter(
                _["reportEntryConstraintMessageTemplate"]["constraintName"]
                for _ in resource_violations
            )
            test_case_name = f"{resource}"
            if violations_count == 0:
                test_case_name += " has no violations"
            if violations_count == 1:
                test_case_name += f" has 1 violation ({next(iter(constraints.keys()))})"
            if violations_count > 1:
                constrains_str = ""
                for constraint, constraint_count in constraints.items():
                    constrains_str += f", {constraint_count}x{constraint}"
                test_case_name += f" has {violations_count} violations ({constrains_str[2:]})"

            test_case = TestCase(
                name=test_case_name,
                classname=f"{context_graph} tested with {shape_graph}",
                elapsed_sec=average_elapsed_sec,
            )
            if violations_count > 0:
                test_case.add_failure_info(output=json.dumps(resource_violations, indent=2))
            test_cases.append(test_case)
        test_suite = TestSuite(
            name=f"Testing {context_graph} with shapes from {shape_graph}.",
            test_cases=test_cases,
            id=report["id"],
            timestamp=report["executionFinished"],
        )
        test_suites.append(test_suite)
    return str(to_xml_report_string(test_suites, encoding="utf-8"))


def get_sorted_validations_list(ctx: Context) -> list[dict]:  # noqa: ARG001
    """Get a sorted list of validation objects (aggregations)"""
    objects = validation.get_all_aggregations()
    return sorted(objects, key=lambda o: str(o.get("executionStarted", "SCHEDULED")))


validations_list = ObjectList(
    name="validation processes",
    get_objects=get_sorted_validations_list,
    filters=[
        DirectValuePropertyFilter(
            name="status",
            description="Filter list by current status of the process.",
            property_key="state",
            transform=transform_lower,
        ),
        DirectValuePropertyFilter(
            name="context-graph",
            description="Filter list by used data / context graph IRI.",
            property_key="contextGraphIri",
        ),
        DirectValuePropertyFilter(
            name="shape-graph",
            description="Filter list by used shape graph IRI.",
            property_key="shapeGraphIri",
        ),
        DirectValuePropertyFilter(
            name="more-resources-than",
            description="Filter list by the number of resources.",
            property_key="resourceCount",
            compare=compare_int_greater_than,
            fixed_completion=[
                CompletionItem("0"),
                CompletionItem("100"),
                CompletionItem("500"),
            ],
        ),
        DirectValuePropertyFilter(
            name="more-violations-than",
            description="Filter list by the number of violations.",
            property_key="violationsCount",
            compare=compare_int_greater_than,
            fixed_completion=[
                CompletionItem("0"),
                CompletionItem("100"),
                CompletionItem("500"),
            ],
        ),
        DirectValuePropertyFilter(
            name="more-violated-resources-than",
            description="Filter list by the number of violated resources.",
            property_key="resourcesWithViolationsCount",
            compare=compare_int_greater_than,
            fixed_completion=[
                CompletionItem("0"),
                CompletionItem("100"),
                CompletionItem("500"),
            ],
        ),
    ],
)


def get_violations_list(ctx: Context) -> list[dict]:
    """Get a list of violations"""
    try:
        # sometimes process_id is in params, sometimes in args !?
        process_id = ctx.params.get("process_id", None)
        if not process_id:
            process_id = ctx.args[0]
    except IndexError:
        return []  # process_id not given
    violations: list[dict] = []  # create a new object which better matches object_list needs
    for result in validation.get(batch_id=process_id)["results"]:
        resource_iri = result["resourceIri"]
        node_shapes = result["nodeShapes"]
        for _ in result["violations"]:
            _["resourceIri"] = resource_iri
            _["nodeShapes"] = node_shapes
            _["constraintName"] = _["reportEntryConstraintMessageTemplate"]["constraintName"]
            violations.append(_)
    return violations


violations_list = ObjectList(
    name="violations",
    get_objects=get_violations_list,
    filters=[
        DirectValuePropertyFilter(
            name="constraint",
            description="Filter list by constraint name.",
            property_key="constraintName",
        ),
        DirectValuePropertyFilter(
            name="severity", description="Filter list by severity.", property_key="severity"
        ),
        DirectValuePropertyFilter(
            name="resource",
            description="Filter list by resource IRI.",
            property_key="resourceIri",
            title_helper=TitleHelper(),
        ),
        DirectListPropertyFilter(
            name="node-shape",
            description="Filter list by node shape IRI.",
            property_key="nodeShapes",
            title_helper=TitleHelper(),
        ),
        DirectValuePropertyFilter(
            name="source",
            description="Filter list by constraint source.",
            property_key="source",
            title_helper=TitleHelper(),
        ),
    ],
)


def _get_batch_validation_option(validation_: dict) -> tuple[str, str]:
    """Get a completion option of a single batch validation"""
    id_ = validation_["id"]
    state = validation_["state"]
    graph = validation_["contextGraphIri"]
    stamp = datetime.fromtimestamp(validation_["executionStarted"] / 1000, tz=timezone.utc)
    time_ago = timeago.format(stamp, datetime.now(tz=timezone.utc))
    resources = _get_resource_count(validation_)
    violations = _get_violation_count(validation_)
    return (
        id_,
        f"{state} - {time_ago}, {resources} resources, {violations} violations ({graph})",
    )


def _complete_all_batch_validations(
    ctx: click.Context,  # noqa: ARG001
    param: click.Argument,  # noqa: ARG001
    incomplete: str,
) -> list[CompletionItem]:
    """Provide completion for batch validation"""
    options = [_get_batch_validation_option(_) for _ in validation.get_all_aggregations()]
    return finalize_completion(candidates=options, incomplete=incomplete)


def _complete_running_batch_validations(
    ctx: click.Context,  # noqa: ARG001
    param: click.Argument,  # noqa: ARG001
    incomplete: str,
) -> list[CompletionItem]:
    """Provide completion for running batch validation"""
    options = [
        _get_batch_validation_option(_)
        for _ in validation.get_all_aggregations()
        if _["state"] == validation.STATUS_RUNNING
    ]
    return finalize_completion(candidates=options, incomplete=incomplete)


def _complete_finished_batch_validations(
    ctx: click.Context,  # noqa: ARG001
    param: click.Argument,  # noqa: ARG001
    incomplete: str,
) -> list[CompletionItem]:
    """Provide completion for finished batch validations"""
    options = [
        _get_batch_validation_option(_)
        for _ in validation.get_all_aggregations()
        if _["state"] == validation.STATUS_FINISHED
    ]
    return finalize_completion(candidates=options, incomplete=incomplete)


def _print_process_summary(app: ApplicationContext, process_id: str) -> None:
    """Show summary of the validation process"""
    app.echo_info_table(
        struct_to_table(validation.get_aggregation(batch_id=process_id)),
        headers=["Key", "Value"],
        sort_column=0,
        caption="Validation Summary",
    )


def show_violated_resources(app: ApplicationContext, data: list[dict]) -> None:
    """Show violated resource IRIs of a validation process"""
    messages = sorted({_["resourceIri"] for _ in data})
    app.echo_info(message=messages)


def _wait_for_process_completion(
    app: ApplicationContext, process_id: str, polling_interval: int, use_rich: bool = False
) -> str:
    class State:
        """State of a validation process"""

        id_: str
        data: dict
        status: str
        completed: int
        total: int

        def __init__(self, id_: str):
            self.id_ = id_
            self.refresh()

        def refresh(self) -> None:
            self.data = validation.get_aggregation(batch_id=self.id_)
            self.status = self.data.get("state", "UNKNOWN")
            self.completed = self.data.get("resourceProcessedCount", 0)
            self.total = self.data.get("resourceCount", 0)
            app.echo_debug(f"Process {self.id_} has status {self.status}.")

    state = State(id_=process_id)
    progress: Progress | None = None
    task: TaskID | None = None
    if use_rich:
        progress = Progress(
            SpinnerColumn(),
            *Progress.get_default_columns(),
            TimeElapsedColumn(),
            transient=True,
            console=app.console,
        )
        progress.__enter__()  # simulate context manager (with:)
        task = progress.add_task(f"{state.status.capitalize()} ... ", total=state.total)
    while True:
        time.sleep(polling_interval)
        state.refresh()
        if progress is not None and task is not None:
            progress.update(
                task_id=task,
                completed=state.completed,
                description=f"{state.status.capitalize()} ... {state.completed} / {state.total}",
            )
        if state.status in (validation.STATUS_SCHEDULED, validation.STATUS_RUNNING):
            # when reported as running or scheduled, start another loop
            continue
        # when reported as finished, error or cancelled break out
        break
    if progress is not None and task is not None:
        progress.stop()
        progress.__exit__(None, None, None)
    if state.status == validation.STATUS_CANCELLED:
        raise ServerError(app, "Process was cancelled.")
    if state.status == validation.STATUS_ERROR:
        error_message = state.data.get("error", "")
        raise ServerError(app, f"Process ended with error: {error_message}")
    return state.status


def _print_violation_table(
    app: ApplicationContext, data_graph: str, shape_graph: str, violations: list[dict]
) -> None:
    """Print violation table from batch validation result"""
    # fetch titles
    resources = []
    for violation in violations:
        resources.append(str(violation.get("resourceIri")))
        resources.extend(violation.get("nodeShapes", []))
    title_helper = TitleHelper()
    title_helper.get(resources)

    # prepare link helper
    resource_link = ResourceLink(graph_iri=data_graph, title_helper=title_helper)
    shape_link = ResourceLink(graph_iri=shape_graph, title_helper=title_helper)

    table = []
    for violation in violations:
        combined_cell = ""

        path = violation.get("path", None)
        if path is not None:
            combined_cell = f"Path: {path}\n"

        source = violation.get("source", None)
        if source is not None:
            combined_cell = f"{combined_cell}Source: {shape_link.process(text=source)}"

        node_shapes = violation.get("nodeShapes", [])
        if len(node_shapes) == 1:
            combined_cell = f"{combined_cell}\nNodeShape: {shape_link.process(text=node_shapes[0])}"
        if len(node_shapes) > 1:
            combined_cell = f"{combined_cell}\nNodeShapes:"
            for node_shape in node_shapes:
                combined_cell = f"{combined_cell}\n - {shape_link.process(text=node_shape)}"

        text = violation["messages"][0]["value"]  # default: use the text of the first message
        for message in violation["messages"]:
            # look for en non non-lang messages to use
            if message["lang"] == "" or message["lang"] == "en":
                text = str(message["value"])
                break
        combined_cell = f"{combined_cell}\nMessage: {text}"

        row = [
            resource_link.process(text=str(violation.get("resourceIri"))),
            violation.get("constraintName", "UNKNOWN"),
            combined_cell,
        ]
        table.append(row)

    app.echo_info_table(
        table,
        headers=["Resource", "Constraint", "Details"],
        sort_column=0,
        caption="Violation List",
        empty_table_message="No violations found.",
    )


def _get_resource_count(batch_validation: dict) -> str:
    """Get resource count from validation report"""
    resource_count = str(batch_validation.get("resourceCount", "-"))
    processed_count = str(batch_validation.get("resourceProcessedCount", "-"))
    if resource_count == processed_count:
        return resource_count
    return f"{processed_count} / {resource_count}"


def _get_violation_count(process_data: dict) -> str:
    """Get violation count from validation report"""
    if process_data.get("executionStarted") is None:
        return "-"
    resources = str(process_data.get("resourcesWithViolationsCount", "0"))
    violations = str(process_data.get("violationsCount", "0"))
    if violations == "0":
        return "0"
    return f"{violations} ({resources} Resources)"


@click.command(cls=CmemcCommand, name="execute")
@click.argument("iri", type=click.STRING, shell_complete=completion.graph_uris)
@click.option(
    "--wait",
    is_flag=True,
    help="Wait until the process is finished. When using this option without the "
    "`--id-only` flag, it will enable a progress bar and a summary view.",
)
@click.option(
    "--shape-graph",
    shell_complete=completion.graph_uris_skip_check,
    default="https://vocab.eccenca.com/shacl/",
    show_default=True,
    help="The shape catalog used for validation.",
)
@click.option(
    "--query",
    shell_complete=completion.remote_queries_and_sparql_files,
    help="SPARQL query to select the resources which you want to validate from "
    "the data graph. "
    "Can be provided as a local file or as a query catalog IRI. "
    "[default: all typed resources]",
)
@click.option(
    "--result-graph",
    shell_complete=completion.writable_graph_uris,
    help="(Optionally) write the validation results to a Knowledge Graph. " "[default: None]",
)
@click.option(
    "--replace",
    is_flag=True,
    default=False,
    help="Replace the result graph instead of just adding the new results. "
    "This is a dangerous option, so use it with care!",
)
@click.option(
    "--ignore-graph",
    shell_complete=completion.ignore_graph_uris,
    type=click.STRING,
    multiple=True,
    help="A set of data graph IRIs which are not queried in the resource selection. "
    "This option is useful for validating only parts of an integration graph "
    "which imports other graphs.",
)
@click.option(
    "--id-only",
    is_flag=True,
    help="Return the validation process identifier only. "
    "This is useful for piping the ID into other commands.",
)
@click.option(
    "--inspect",
    is_flag=True,
    help="Return the list of violations instead of the summary (includes --wait).",
)
@click.option(
    "--polling-interval",
    type=click.IntRange(min=1),
    show_default=True,
    default=1,
    help="How many seconds to wait between status polls. Status polls are"
    " cheap, so a higher polling interval is most likely not needed.",
)
@click.pass_context
def execute_command(  # noqa: PLR0913
    ctx: Context,
    iri: str,
    shape_graph: str,
    query: str,
    result_graph: str,
    replace: bool,
    ignore_graph: list[str],
    id_only: bool,
    wait: bool,
    inspect: bool,
    polling_interval: int,
) -> None:
    """Start a new validation process.

    Validation is performed on all typed resources of the data / context graph
    (and its sub-graphs). Each resource is validated against all applicable node
    shapes from the shape catalog.
    """
    app: ApplicationContext = ctx.obj
    if id_only and inspect:
        raise UsageError(
            "Output can be the summary (default), the process ID (--id-only) "
            "or the violation list (--inspect)."
        )
    process_id = validation.start(
        context_graph=iri,
        shape_graph=shape_graph,
        query=get_query_text(query, {"resource"}) if query else None,
        result_graph=result_graph,
        replace=replace,
        ignore_graph=ignore_graph,
    )
    if wait or inspect:
        _wait_for_process_completion(
            app=app, process_id=process_id, use_rich=not id_only, polling_interval=polling_interval
        )
    if id_only:
        app.echo_info(process_id)
        return
    if inspect:
        ctx.params["process_id"] = process_id
        data = violations_list.apply_filters(ctx=ctx, filter_=[])
        _print_violation_table(app=app, data_graph=iri, shape_graph=shape_graph, violations=data)
        return
    _print_process_summary(process_id=process_id, app=app)


@click.command(cls=CmemcCommand, name="list")
@click.option(
    "--filter",
    "filter_",
    type=(str, str),
    multiple=True,
    help=validations_list.get_filter_help_text(),
    shell_complete=validations_list.complete_values,
)
@click.option(
    "--id-only",
    is_flag=True,
    help="List validation process identifier only. "
    "This is useful for piping the IDs into other commands.",
)
@click.option("--raw", is_flag=True, help="Outputs raw JSON of the validation list.")
@click.pass_context
def list_command(ctx: Context, filter_: tuple[tuple[str, str]], id_only: bool, raw: bool) -> None:
    """List running and finished validation processes.

    This command provides a filterable table or identifier list of validation
    processes. The command operates on the process summary and provides some statistics.

    Note: Detailed information on the found violations can be listed with the
    `graph validation inspect` command.
    """
    validations = validations_list.apply_filters(ctx=ctx, filter_=filter_)
    app: ApplicationContext = ctx.obj

    if raw:
        app.echo_info_json(validations)
        return

    if id_only:
        for _ in validations:
            app.echo_info(_["id"])
        return

    # output a user table
    table = []
    for _ in validations:
        row = [
            _["id"],
            _["state"],
            _.get("executionStarted", None),
            _["contextGraphIri"],
            _get_resource_count(_),
            _get_violation_count(_),
        ]
        table.append(row)
    app.echo_info_table(
        table,
        headers=["ID", "Status", "Started", "Graph", "Resources", "Violations"],
        caption=f"Validation Processes of {get_cmem_base_uri()}",
        cell_processing={2: TimeAgo(), 3: GraphLink()},
        empty_table_message="No validation processes found. "
        "Use `graph validation execute` to start a new validation process.",
    )


@click.command(cls=CmemcCommand, name="inspect")
@click.argument("process_id", type=click.STRING, shell_complete=_complete_all_batch_validations)
@click.option(
    "--filter",
    "filter_",
    type=(str, str),
    multiple=True,
    help=violations_list.get_filter_help_text(),
    shell_complete=violations_list.complete_values,
)
@click.option(
    "--id-only",
    is_flag=True,
    help="Return violated resource identifier only. "
    "This is useful for piping the ID into other commands.",
)
@click.option(
    "--summary",
    is_flag=True,
    help="Outputs the summary of the graph validation "
    "instead of the violations list (not filterable).",
)
@click.option("--raw", is_flag=True, help="Outputs raw JSON of the validation result.")
@click.pass_context
def inspect_command(  # noqa: PLR0913
    ctx: Context,
    process_id: str,
    filter_: tuple[tuple[str, str]],
    id_only: bool,
    summary: bool,
    raw: bool,
) -> None:
    """List and inspect errors found with a validation process.

    This command provides detailed information on the found violations of
    a validation process.

    Use the `--filter` option to limit the output based on different criteria such as
    constraint name (`constraint`), origin node shape of the rule (`node-shape`), or
    the validated resource (`resource`).

    Note: Validation processes IDs can be listed with the `graph validation list`
    command, or by utilizing the tab completion of this command.
    """
    app: ApplicationContext = ctx.obj
    if process_id not in [_["id"] for _ in validation.get_all_aggregations()]:
        raise UsageError(f"Validation process with ID '{process_id}' is not known (anymore).")

    if summary:
        if raw:
            app.echo_info_json(validation.get_aggregation(batch_id=process_id))
        else:
            _print_process_summary(app=app, process_id=process_id)
        return

    data = violations_list.apply_filters(ctx=ctx, filter_=filter_)

    if id_only:
        show_violated_resources(app=app, data=data)
        return

    if raw:
        app.echo_info_json(data)
        return

    if len(data) == 0 and len(filter_) == 0:
        app.echo_warning(
            "The given validation process does not have any violations - "
            "I will show the summary instead."
        )
        _print_process_summary(app=app, process_id=process_id)
    else:
        process_data = validation.get_aggregation(batch_id=process_id)
        _print_violation_table(
            app=app,
            violations=data,
            shape_graph=process_data["shapeGraphIri"],
            data_graph=process_data["contextGraphIri"],
        )


@click.command(cls=CmemcCommand, name="cancel")
@click.argument("process_id", type=click.STRING, shell_complete=_complete_running_batch_validations)
@click.pass_obj
def cancel_command(app: ApplicationContext, process_id: str) -> None:
    """Cancel a running validation process.

    Note: In order to get the process IDs of all currently running validation
    processes, use the `graph validation list` command with the option
    `--filter status running`, or utilize the tab completion of this command.
    """
    all_status = {_["id"]: _["state"] for _ in validation.get_all_aggregations()}
    if process_id not in all_status:
        raise UsageError(f"Validation process with ID '{process_id}' is not known (anymore).")
    if all_status[process_id] != validation.STATUS_RUNNING:
        raise click.UsageError(
            f"Validation process with ID '{process_id}' is not a running anymore."
        )
    app.echo_info(f"Validation process with ID '{process_id}' ... ", nl=False)
    validation.cancel(batch_id=process_id)
    app.echo_success("cancelled")


@click.command(cls=CmemcCommand, name="export")
@click.argument(
    "process_ids", nargs=-1, type=click.STRING, shell_complete=_complete_finished_batch_validations
)
@click.option(
    "--output-file",
    type=click.Path(writable=True, allow_dash=False, dir_okay=False),
    default="report.xml",
    show_default=True,
    help="Export the report to this file. Existing files will be overwritten.",
)
@click.option(
    "--exit-1",
    type=click.Choice(["never", "error"]),
    default="error",
    show_default=True,
    help="Specify, when this command returns with exit code 1. Available options are "
    "'never' (exit 0, even if there are violations in the reports), "
    "'error' (exit 1 if there is at least one violation in a report).), ",
)
@click.option(
    "--format",
    "format_",
    type=click.Choice(["JSON", "XML"], case_sensitive=True),
    default="XML",
    help="Export either the plain JSON report or a distilled jUnit XML report.",
    show_default=True,
)
@click.pass_context
def export_command(
    ctx: Context, process_ids: tuple[str], output_file: str, exit_1: str, format_: str
) -> None:
    """Export a report of finished validations.

    This command exports a jUnit XML or JSON report in order to process
    them somewhere else (e.g. a CI pipeline).

    You can export a single report of multiple validation processes.

    For jUnit XML: Each validation process result will be transformed to
    a single test suite. All violations of one resource in a result will be
    collected and attached to a single test case in that test suite.

    Note: Validation processes IDs can be listed with the `graph validation list`
    command, or by utilizing the tab completion of this command.
    """
    if len(process_ids) == 0:
        raise UsageError("This command needs at least one validation process ID.")
    app: ApplicationContext = ctx.obj
    process_ids_to_test = {_: True for _ in process_ids}
    overall_violations = 0
    overall_resources = 0
    for _ in validation.get_all_aggregations():
        if _["id"] in process_ids_to_test:
            if _["state"] != "FINISHED":
                raise UsageError(f"Validation process with ID '{_['id']}' is still running.")
            del process_ids_to_test[_["id"]]
            overall_violations += int(_["violationsCount"])
            overall_resources += int(_["resourceProcessedCount"])
    if len(process_ids_to_test) > 0:
        raise UsageError(
            "Validation processes with the following IDs not known (anymore): "
            + ", ".join(process_ids_to_test)
        )
    reports = []
    for process_id in process_ids:
        report = validation.get(batch_id=process_id)
        reports.append(report)
    app.echo_info(
        f"Export of {len(reports)} validation report(s) with"
        f" {overall_violations} violations in {overall_resources} resources"
        f" to {output_file} ... ",
        nl=False,
    )
    with Path(output_file, mode="w", encoding="utf-8") as file:
        if format_ == "XML":
            file.write_text(_reports_to_junit(reports))
        if format_ == "JSON":
            file.write_text(json.dumps(reports, indent=2))
    app.echo_success("done")
    if exit_1 == "error" and overall_violations > 0:
        app.echo_error(
            "Exit 1 since violations where found in the reports "
            "(can be suppressed with '--exit-1 never')."
        )
        sys.exit(1)


@click.group(cls=CmemcGroup, name="validation")
def validation_group() -> CmemcGroup:  # type: ignore[empty-body]
    """Validate resources in a graph.

    This command group is dedicated to the management of resource validation processes.
    A validation process verifies, that resources in a specific graph are valid according
    to the node shapes in a shape catalog graph.

    Note: Validation processes are identified with a random ID and can be listed with
    the `graph validation list` command. To start or cancel validation processes,
    use the `graph validation execute` and `graph validation cancel` command.
    To inspect the found violations of a validation process, use the
    `graph validation inspect` command.
    """


validation_group.add_command(execute_command)
validation_group.add_command(list_command)
validation_group.add_command(inspect_command)
validation_group.add_command(cancel_command)
validation_group.add_command(export_command)



"""The main command line interface."""

import ast
import configparser
import json
import os
import re
import subprocess  # nosec
from datetime import datetime, timezone
from os import environ as env
from os import getenv
from pathlib import Path
from shutil import which

import click
import cmem.cmempy.config as cmempy_config
import urllib3
from cmem.cmempy.health import get_di_version, get_explore_version
from pygments import highlight
from pygments.formatters import get_formatter_by_name
from pygments.lexers import get_lexer_by_name
from rich import box
from rich.console import Console
from rich.table import Table
from urllib3.exceptions import InsecureRequestWarning

from cmem_cmemc.config_parser import PureSectionConfigParser
from cmem_cmemc.exceptions import InvalidConfigurationError
from cmem_cmemc.string_processor import StringProcessor, process_row
from cmem_cmemc.utils import is_enabled, str_to_bool

DI_TARGET_VERSION = "v25.1.0"

EXPLORE_TARGET_VERSION = "v25.1.0"

KNOWN_CONFIG_KEYS = {
    "CMEM_BASE_URI": cmempy_config.get_cmem_base_uri,
    "SSL_VERIFY": cmempy_config.get_ssl_verify,
    "REQUESTS_CA_BUNDLE": cmempy_config.get_requests_ca_bundle,
    "DP_API_ENDPOINT": cmempy_config.get_dp_api_endpoint,
    "DI_API_ENDPOINT": cmempy_config.get_di_api_endpoint,
    "KEYCLOAK_BASE_URI": cmempy_config.get_keycloak_base_uri,
    "KEYCLOAK_REALM_ID": cmempy_config.get_keycloak_realm_id,
    "OAUTH_TOKEN_URI": cmempy_config.get_oauth_token_uri,
    "OAUTH_GRANT_TYPE": cmempy_config.get_oauth_grant_type,
    "OAUTH_USER": cmempy_config.get_oauth_user,
    "OAUTH_PASSWORD": cmempy_config.get_oauth_password,
    "OAUTH_CLIENT_ID": cmempy_config.get_oauth_client_id,
    "OAUTH_CLIENT_SECRET": cmempy_config.get_oauth_client_secret,
    "OAUTH_ACCESS_TOKEN": cmempy_config.get_oauth_access_token,
}

KNOWN_SECRET_KEYS = ("OAUTH_PASSWORD", "OAUTH_CLIENT_SECRET", "OAUTH_ACCESS_TOKEN")

SSL_VERIFY_WARNING = "SSL verification is disabled (SSL_VERIFY=False)."


class ApplicationContext:
    """Context of the command line interface."""

    APP_NAME: str = "cmemc"
    DEFAULT_CONFIG_FILE: str = str(Path(click.get_app_dir(APP_NAME)) / "config.ini")
    DEFAULT_EXTERNAL_HTTP_TIMEOUT: int = 10

    debug: bool
    quiet: bool
    external_http_timeout: int
    config_dir: Path
    config_file: Path
    config: PureSectionConfigParser
    connection: str | None
    console: Console
    console_width: int | None = None

    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-many-public-methods

    def __init__(
        self,
        config_file: str,
        connection: str | None = None,
        debug: bool = False,
        quiet: bool = False,
    ):
        """Initialize main context."""
        self.config_file = Path(config_file)
        self.app_name = "cmemc"
        self.set_debug(debug)
        self.set_quiet(quiet)
        self.ensure_app_config_dir()
        self.set_connection(connection)
        self.set_external_http_timeout(self.DEFAULT_EXTERNAL_HTTP_TIMEOUT)
        self.console = Console(markup=True, emoji_variant="emoji")
        self.update_console_width()

    @staticmethod
    def from_params(params: dict) -> "ApplicationContext":
        """Create an ApplicationContext instance from a dictionary.

        Expects 'config_file' key to be present in the dictionary.
        """
        config_file = params.get("config_file")
        debug = str_to_bool(str(params.get("debug")))
        quiet = str_to_bool(str(params.get("quiet")))
        external_http_timeout = (
            int(str(params.get("external_http_timeout")))
            if "external_http_timeout" in params
            else ApplicationContext.DEFAULT_EXTERNAL_HTTP_TIMEOUT
        )
        connection = str(params.get("connection")) if params.get("connection") else None
        if not config_file:
            raise ValueError("Missing required key: 'config_file' in config dictionary")
        app = ApplicationContext(
            config_file=config_file, connection=connection, debug=debug, quiet=quiet
        )
        app.set_external_http_timeout(external_http_timeout)
        if not app.debug:
            app.set_debug(is_enabled(params, app.get_config(), "debug"))
        if not app.quiet:
            app.set_quiet(is_enabled(params, app.get_config(), "quiet"))
        app.echo_debug(f"use config file: {app.config_file}")
        if app.connection:
            app.echo_debug(f"set connection: {app.connection}")
        defaults = app.get_defaults()
        app.apply_config(defaults)
        section_config = app.get_connection_config()
        app.clear_known_keys()
        app.apply_config(section_config)
        app.configure_cmempy()
        return app

    def update_console_width(self) -> None:
        """Update console width from environment variable."""
        console_width_env = getenv("CMEMC_CONSOLE_WIDTH", None)
        console_width = int(console_width_env) if isinstance(console_width_env, str) else None
        self.console_width = console_width
        if console_width is not None:
            self.console.width = console_width

    def get_template_data(self) -> dict[str, str]:
        """Get the template data dict with vars from the context."""
        data: dict[str, str] = {}
        today = str(datetime.now(tz=timezone.utc).date())
        data.update(date=today)
        if self.connection is not None:
            data.update(connection=self.connection)
        else:
            data.update(connection="unnamed")
        return data

    def set_debug(self, debug: bool = False) -> None:
        """Set debug state"""
        self.debug = debug

    def set_quiet(self, quiet: bool = False) -> None:
        """Set quiets state"""
        self.quiet = quiet

    def set_external_http_timeout(self, timeout: int) -> None:
        """Set external http timeout"""
        self.external_http_timeout = timeout
        os.environ["CMEMC_EXTERNAL_HTTP_TIMEOUT"] = str(timeout)

    def ensure_app_config_dir(self) -> None:
        """Ensure the application's configuration directory exists."""
        try:
            self.config_dir = Path(click.get_app_dir(self.app_name))
            self.config_dir.mkdir(parents=True, exist_ok=True)
        except (OSError, PermissionError, FileNotFoundError):
            self.echo_debug(f"Could not create config directory {self.config_dir}")

    def set_config_file(self, config_file: str | None = None) -> None:
        """Set and return the context config file"""
        if config_file:
            self.config_file = Path(config_file)
        else:
            self.config_file = self.config_dir / "config.ini"
        self.echo_debug(f"Set config to {self.config_file.absolute()}")

    def get_defaults(self) -> dict[str, str]:
        """Populate missing values in `params` using defaults from the configuration."""
        config = self.get_config()
        defaults = dict(config.defaults())
        if "CMEMC_CONFIG_FILE" in defaults:
            defaults.pop("CMEMC_CONFIG_FILE")
        if "CMEMC_DEBUG" in defaults and self.debug:
            defaults.pop("CMEMC_DEBUG")
        if "CMEMC_QUIET" in defaults and self.quiet:
            defaults.pop("CMEMC_QUIET")
        if "CMEMC_CONNECTION" in defaults and self.connection:
            defaults.pop("CMEMC_CONNECTION")
        if (
            "CMEMC_EXTERNAL_HTTP_TIMEOUT" in defaults
            and self.external_http_timeout != self.DEFAULT_EXTERNAL_HTTP_TIMEOUT
        ):
            defaults.pop("CMEMC_EXTERNAL_HTTP_TIMEOUT")
        for key in defaults:
            if key in os.environ:
                defaults.pop(key)
        return defaults

    def get_connection_config(self) -> dict[str, str]:
        """Get connection section config"""
        if not self.connection:
            return {}
        config = dict(self.get_config()[self.connection])
        if "CMEMC_CONFIG_FILE" in config:
            config.pop("CMEMC_CONFIG_FILE")
        if "CMEMC_CONNECTION" in config:
            config.pop("CMEMC_CONNECTION")
        return config

    def apply_config(self, config: dict[str, str]) -> None:
        """Apply the configuration to the context."""
        for key, value in config.items():
            if key == "CMEMC_DEBUG":
                self.set_debug(str_to_bool(value))
            elif key == "CMEMC_QUIET":
                self.set_quiet(str_to_bool(value))
            elif key == "CMEMC_CONNECTION":
                self.echo_debug(f"set connection to {value}")
                self.set_connection(value)
            elif key == "CMEMC_EXTERNAL_HTTP_TIMEOUT":
                self.set_external_http_timeout(int(value))
            else:
                if key in KNOWN_SECRET_KEYS:
                    self.echo_debug(key + " set by config")
                else:
                    self.echo_debug(key + " set by config to " + value)
                env[key] = value

    def clear_known_keys(self) -> None:
        """Clear know keys if the connection is set."""
        if self.connection:
            # clear all known keys
            for key in KNOWN_CONFIG_KEYS:
                if key in env:
                    env.pop(key)

    def configure_cmempy(self) -> None:
        """Configure the cmempy API to use a new connection."""
        # With or without connection config, we do not want API stdout prints
        env["CMEMPY_IS_CHATTY"] = "False"
        if self.connection:
            config = self.get_config()[self.connection]
            # allow to fetch all secrets from an external process
            for _ in KNOWN_SECRET_KEYS:
                self.set_credential_from_process(_, _ + "_PROCESS", config)

            self.echo_debug(f"CA bundle loaded from {cmempy_config.get_requests_ca_bundle()}")
        # If cert validation is disabled, output a warning
        # Also disable library warnings:
        # https://urllib3.readthedocs.io/en/latest/advanced-usage.html#tls-warnings
        if not cmempy_config.get_ssl_verify():
            self.echo_warning(SSL_VERIFY_WARNING)
            urllib3.disable_warnings(category=InsecureRequestWarning)

    @staticmethod
    def set_connection_from_params(params: dict) -> None:
        """Set connection and config by manually checking params (completion)."""
        app = ApplicationContext.from_params(params)
        defaults = app.get_defaults()
        app.apply_config(defaults)
        section_config = app.get_connection_config()
        app.clear_known_keys()
        app.apply_config(section_config)
        app.configure_cmempy()

    def set_connection(self, section_string: str | None = None) -> None:
        """Set connection config section based on section string."""
        self.connection = None
        if section_string is None or section_string == "":
            return
        if section_string not in self.get_config():
            raise InvalidConfigurationError(
                self,
                f"There is no connection '{section_string}' configured in "
                f"config '{self.config_file}'.",
            )
        self.connection = section_string

    def get_config_file(self) -> Path:
        """Check the connection config file."""
        if not self.config_file.exists():
            with self.config_file.open(mode="a", encoding="UTF-8"):
                self.echo_warning(f"Empty config created: {self.config_file.name}")
        return self.config_file

    def get_config(self) -> PureSectionConfigParser:
        """Parse the configuration"""
        config = PureSectionConfigParser()
        try:
            # https://stackoverflow.com/questions/1648517/
            config.read(self.get_config_file(), encoding="utf-8")
        except configparser.Error as error:
            raise InvalidConfigurationError(
                self,
                "The following config parser error needs to be fixed with your config file:\n"
                f"{error!s}\n"
                "You can use the 'config edit' command to fix this.",
            ) from error
        except Exception as error:  # noqa: BLE001
            self.echo_debug(f"Could not read config file - provide empty config: {error!s}")
        return config

    @staticmethod
    def is_completing() -> bool:
        """Test for environment which indicates that we are in completion mode.

        Returns true if in validation mode, otherwise false.

        Note: moved from utils due to circular imports

        Returns: boolean
        """
        comp_words = os.getenv("COMP_WORDS", default=None)
        cmemc_complete = os.getenv("_CMEMC_COMPLETE", default=None)
        if comp_words is not None:
            return True
        return cmemc_complete is not None

    @staticmethod
    def echo_warning(message: str, nl: bool = True, condition: bool = True) -> None:
        """Output a warning message."""
        if not condition:
            return
        if ApplicationContext.is_completing():
            return
        click.secho(message, fg="yellow", err=True, nl=nl)

    @staticmethod
    def echo_error(
        message: str | list[str], nl: bool = True, err: bool = True, prepend_line: bool = False
    ) -> None:
        """Output an error message.

        2024-05-17: also allows list of strings now
        2024-05-17: new prepend_line parameter
        """
        # pylint: disable=invalid-name
        click.echo("") if prepend_line is True else None
        messages: list[str] = [message] if isinstance(message, str) else message
        for _ in messages:
            click.secho(_, fg="red", err=err, nl=nl)

    def echo_debug(self, message: str | list[str]) -> None:
        """Output a debug message if --debug is enabled.

        2024-05-17: also allows list of strings now
        2025-04-07: extract output function
        """
        messages: list[str] = [message] if isinstance(message, str) else message
        if self.debug and not ApplicationContext.is_completing():
            self.echo_debug_string(messages)

    @staticmethod
    def echo_debug_string(message: str | list[str]) -> None:
        """Output a debug message"""
        messages: list[str] = [message] if isinstance(message, str) else message
        now = datetime.now(tz=timezone.utc)
        for _ in messages:
            click.secho(f"[{now!s}] {_}", err=True, dim=True)

    def echo_info(
        self,
        message: str | list[str] | set[str],
        nl: bool = True,
        fg: str = "",
        condition: bool = True,
    ) -> None:
        """Output one or more info messages, if not suppressed by --quiet.

        message: the string to output
        nl: True if newlines need to be added (default)
        fg: color
        condition: do nothing if False
        """
        if not condition:
            return
        if self.quiet:
            return
        if isinstance(message, str):
            click.secho(message, nl=nl, fg=fg)
            return
        if isinstance(message, list | set):
            for msg in message:
                click.secho(msg, nl=nl, fg=fg)

    def echo_info_json(self, object_: object) -> None:
        """Output a formatted and highlighted json as info message."""
        message = highlight(
            json.dumps(object_, indent=2),
            get_lexer_by_name("json"),
            get_formatter_by_name("terminal"),
        )
        self.echo_info(message)

    def echo_info_xml(self, document: str) -> None:
        """Output a formatted and highlighted XML as info message."""
        message = highlight(
            document,
            get_lexer_by_name("xml"),
            get_formatter_by_name("terminal"),
        )
        self.echo_info(message)

    def echo_info_table(  # noqa: PLR0913
        self,
        rows: list,
        headers: list[str],
        sort_column: int | None = None,
        caption: str | None = None,
        cell_processing: dict[int, StringProcessor] | None = None,
        empty_table_message: str | None = None,
    ) -> None:
        """Output a formatted and highlighted table as info message."""
        self.update_console_width()

        if len(rows) == 0 and empty_table_message:
            self.echo_warning(empty_table_message)
            return

        if sort_column is not None:
            rows = sorted(rows, key=lambda k: k[sort_column].lower())

        if cell_processing:
            rows = [process_row(row, cell_processing) for row in rows]

        table = Table(
            box=box.HEAVY,
            row_styles=["bold", ""],
            header_style="Yellow",
            border_style="Yellow",
        )
        if caption is not None:
            table.caption = caption
        for header in headers:
            table.add_column(header, overflow="fold")
        for row_source in rows:
            row = [str(cell) for cell in row_source]
            table.add_row(*row)
        self.console.print(table)

    def echo_info_sparql(self, query_text: str) -> None:
        """Output a formatted and highlighted sparql query as info message."""
        message = highlight(
            query_text, get_lexer_by_name("sparql"), get_formatter_by_name("terminal")
        )
        self.echo_info(message)

    def echo_success(self, message: str, nl: bool = True, condition: bool = True) -> None:
        """Output success message, if not suppressed by --quiet."""
        # pylint: disable=invalid-name
        self.echo_info(message, fg="green", nl=nl, condition=condition)

    @staticmethod
    def echo_result(message: str, nl: bool = True) -> None:
        """Output result message, can NOT be suppressed by --quiet."""
        # pylint: disable=invalid-name
        click.echo(message, nl=nl)

    def check_concrete_version(self, name: str, version: str, target_version: str) -> None:
        """Check and compare a concrete backend version against a concrete target version.

        Args:
        ----
            name: Name of the backend
            version: Version of the backend
            target_version: Target version of cmemc for this backend

        """
        # backend version
        if not (match := re.match(r"^v([0-9]+)\.([0-9]+)(.*)?$", version)):
            self.echo_warning(f"There was an error checking the {name} version.")
            return
        year = int(match.group(1))
        major = int(match.group(2))
        # target version
        if not (target_match := re.match(r"^v([0-9]+)\.([0-9]+)(.*)?$", target_version)):
            self.echo_warning(f"There was an error checking the {name} version.")
            return
        target_year = int(target_match.group(1))
        target_major = int(target_match.group(2))
        if year < target_year or (year == target_year and major < target_major):
            self.echo_warning(
                f"Your {name} version v{year}.{major} is lower than the "
                f"target version of your cmemc deployment (v{target_year}.{target_major}).\n"
                "Some feature may be not supported with this backend."
            )

    def check_versions(self) -> None:
        """Check server versions against supported versions."""
        # check Explore version
        self.check_concrete_version(
            name="Explore", version=get_explore_version(), target_version=EXPLORE_TARGET_VERSION
        )
        self.check_concrete_version(
            name="DataIntegration", version=get_di_version(), target_version=DI_TARGET_VERSION
        )

    def set_credential_from_process(
        self, pw_key: str, command_key: str, config: configparser.SectionProxy
    ) -> None:
        """Set a credential env var from the output of an external process.

        Execute a command from the config key command_key and set/overwrite
        the environment variable pw_key with the first line of stdout of
        this process.

        Both stdout and stderr are captured. Output is not shown.

        Args:
        ----
            pw_key (str): the env variable which is created/set
            command_key (str): the env variable which holds the command
            config: the config object

        """
        if command_key not in config:
            return
        self.echo_debug(f"Fetching {pw_key} from external process.")
        command = str(config.get(command_key))
        parsed_list = None
        # first, we try to find command in PATH
        if which(command) is None:
            try:
                # if not in PATH, we try to parse it as list
                parsed_list = list(ast.literal_eval(command))
            except SyntaxError as error:
                raise ValueError(
                    f"'{command}' could not be found in PATH or the value "
                    "could not be parsed as list, e.g. ['echo', 'this']."
                ) from error
            # else: if it is a parsed list, we try to find the first element only
            if which(parsed_list[0]) is None:
                raise ValueError(f"'{parsed_list[0]}' could not be found in PATH.")
            parsed_list[0] = which(parsed_list[0])
        checked_command = parsed_list if parsed_list is not None else command
        if env.get("CMEMC_CREDENTIAL_PROCESS_NO_WARNING", "false") != "true":
            self.echo_warning(
                "Consider possible security implications associated with "
                "executing the external credential process:\n"
                f"{checked_command}\n"
                "You can suppress this warning by setting "
                "CMEMC_CREDENTIAL_PROCESS_NO_WARNING to 'true'."
                ""
            )
        self.echo_debug(f"External credential process started {checked_command}")
        split_output = (
            subprocess.run(  # nosec  # noqa: S603
                checked_command,
                capture_output=True,
                check=True,
            )
            .stdout.decode("utf-8")
            .splitlines()
        )
        if len(split_output) == 0:
            raise ValueError(
                f"The configured credential process '{checked_command}' "
                "did not produce any output on stdout."
            )
        pw_value = split_output[0]
        # set the password environment variable
        env[pw_key] = pw_value
